package edu.bu.LanguageCorrection;

import javax.sql.rowset.spi.SyncFactory;
import javax.swing.*;
import javax.swing.text.Highlighter;
import javax.swing.text.Highlighter.Highlight;

import dotterweide.editor.painter.HighlightPainter;

import java.awt.*;
// import java.awt.event.ActionEvent;
import java.io.ByteArrayOutputStream;
import java.io.PrintStream;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.util.List;
import java.util.Map;
import java.util.ArrayList;
import java.util.HashMap;
import javax.swing.*;
import javax.swing.text.*;
import java.awt.Color;
import java.awt.BorderLayout;
import java.awt.event.ActionEvent;



public class MainApp extends JFrame {
    private final JTextField urlField;
    private JTextArea resultArea;
    private Highlighter.HighlightPainter myHighlightPainter;
    private final JButton runButton;
    private final JComboBox<String> moduleSelector;

    public MainApp() {
        super("Language Correction Tool");
        setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
        setSize(800, 600);
        setLocationRelativeTo(null);
        setLayout(new BorderLayout());

        // Check if metadata file exists
        if (!Files.exists(Paths.get("metadata.ser"))) {
            String[] languages = {"en", "gr", "pt", "it"};
            String selectedLanguage = (String) JOptionPane.showInputDialog(this, "Metadata file not found. Please choose a language to build off of. \n(If you want to build from scratch just click Cancel)", "Language Selection", JOptionPane.PLAIN_MESSAGE, null, languages, languages[0]);
            if (selectedLanguage != null) {
                crawler webCrawler = new crawler();
                webCrawler.build(selectedLanguage);
            }
        }

        // Module selector
        String[] modules = {"Select Module", "Web Crawler", "Reddit Crawler", "File Checker", "File Corrector", "Text Checker", "Text Corrector"};
        moduleSelector = new JComboBox<>(modules);

        // User inout entry field and run button
        urlField = new JTextField();
        runButton = new JButton("Run");
        runButton.addActionListener(e -> {
            String selectedModule = (String) moduleSelector.getSelectedItem();
            String input = urlField.getText().trim();

            switch (selectedModule) {
                case "Web Crawler":
                    runCrawler(input);
                    break;
                case "Reddit Crawler":
                    runCrawler("https://www.reddit.com/r/"+input);
                    break;
                case "File Checker":
                    runChecker(input,true);
                    break;
                case "File Corrector":
                    runCorrector(input,true);
                    break;
                case "Text Checker":
                    runChecker(input,false);
                    break;
                case "Text Corrector":
                    runCorrector(input,false);
                    break;
                default:
                    JOptionPane.showMessageDialog(this, "Select a valid module");
            }
        });
        JPanel northPanel = new JPanel(new BorderLayout());
        northPanel.add(moduleSelector, BorderLayout.WEST);
        northPanel.add(urlField, BorderLayout.CENTER);
        northPanel.add(runButton, BorderLayout.EAST);

        // Result area
        resultArea = new JTextArea();
        resultArea.setEditable(false);
        JScrollPane scrollPane = new JScrollPane(resultArea);

        add(northPanel, BorderLayout.NORTH);
        add(scrollPane, BorderLayout.CENTER);
    }

    private void runCrawler(String input) {
        StringBuilder outputBuilder = new StringBuilder();
        crawler webCrawler = new crawler(output -> {
            outputBuilder.append(output);
            SwingUtilities.invokeLater(() -> resultArea.setText(outputBuilder.toString()));
        });
        if (input.startsWith("http")) {
            // Input is a single URL
            webCrawler.add_to_queue(input);
        } else {
            // Input is file path
            try {
                List<String> lines = Files.readAllLines(Paths.get(input));
                lines.forEach(webCrawler::add_to_queue);
            } catch (Exception e) {
                resultArea.setText("Error reading file: " + e.getMessage());
                return;
            }
        }
        // Create a new thread for running the crawler
        Thread crawlerThread = new Thread(() -> {
            webCrawler.crawl(5); // Adjust the limit (DEFAULT SET TO 1)
        });
        // Start the crawler thread
        crawlerThread.start();
    }

    private void runChecker(String text, boolean isFile) {
        try {
            String content;
            if (isFile) {
                content = new String(Files.readAllBytes(Paths.get(text)));
            } else {
                content = text;
            }

            // Assume content is already properly split into sentences here
            List<String> sentences = TextProcessor.extractSentences(content); // Use a method to split into sentences

            Checker checker = new Checker();
            StringBuilder result = new StringBuilder();

            ByteArrayOutputStream outputStream = new ByteArrayOutputStream();
            PrintStream printStream = new PrintStream(outputStream);
            PrintStream originalOut = System.out;

            String[] worstPhrases = new String[sentences.size()];

            for (String sentence : sentences) {
                System.setOut(printStream);
                checker.analyze(sentence); // Analyze each sentence separately

                // Reset System.out
                System.out.flush();
                System.setOut(originalOut);

                // Capture the output into a string
                String output = outputStream.toString();
                outputStream.reset(); // Clear the output stream for the next sentence
                //System.out.println(output);

                // Parsing the output to get phrases and their scores
                String[] lines = output.split("\n");
                double lowestScore = Double.MAX_VALUE;
                String worstPhrase = null;

                for (String line : lines) {
                    if (line.startsWith("\"phrases\":")) {
                        //System.out.println(line);
                        line = line.replace("\"phrases\":", "").replace("{", "").replace("}", "").trim();
                        //System.out.println(line);
                        String[] phrases = line.split(",");
                        //System.out.println(line);
                        for (String phrase : phrases) {
                            String[] parts = phrase.trim().split(":");
                            double phraseScore = parts[1].trim().equals("null") ? 0
                                    : Double.parseDouble(parts[1].trim());
                            if (phraseScore < lowestScore) {
                                lowestScore = phraseScore;
                                worstPhrase = parts[0].trim();
                                worstPhrases[sentences.indexOf(sentence)] = worstPhrase;
                            }
                        }
                    }
                }

                if (worstPhrase != null) {
                    // Append to the result with annotations
                    result.append("\nSentence: ").append(sentence)
                            .append("\n>> Worst Phrase: ").append(worstPhrase)
                            .append(" (Score: ").append(lowestScore).append(")\n");
                }
            }

            resultArea.setText(result.toString()); // Display the annotated results in the JTextArea

            // Highlight the worst phrase in each sentence
            Highlighter highlighter = resultArea.getHighlighter();
            Highlighter.HighlightPainter painter = new DefaultHighlighter.DefaultHighlightPainter(Color.YELLOW);
            for (String phrase : worstPhrases) {
                phrase = phrase.replaceAll("\"", "");
                //System.out.println(phrase);
                if (phrase != null) {
                    //System.out.println(resultArea.getText());
                    int start = resultArea.getText().indexOf(phrase);
                    if (start != -1) {
                        int end = start + phrase.length();
                        try {
                            highlighter.addHighlight(start, end, painter);
                        } catch (BadLocationException e) {
                            e.printStackTrace();
                        }
                    } else {
                        System.out.println("Text not found");
                    }
                }
            }
        } catch (Exception e) {
            //throw new RuntimeException(e);
            resultArea.setText("Error: " + e.getMessage());
        }
    }

    private void printSentencesInOrderOfChanges(String[] sentences, String originalSentence, StringBuilder result) {
        // Order the sentences by the number of changes needed
        Map<String, Integer> changesMap = new HashMap<>();

        for (String sentence : sentences) {
            if (sentence == null || sentence.isEmpty())
                continue; // Skip empty sentences (not in the 0.5 percentile of the best sentence)

            String sentencePart = sentence.split(" | Score: ")[0]; // Remove score (if present

            int changes = 0;
            if (sentencePart.length() != originalSentence.length()) {
                changes = Math.abs(sentencePart.split(" ").length - originalSentence.split(" ").length) + 1;
            } else {
                String[] originalWords = originalSentence.split(" ");
                String[] correctedWords = sentencePart.split(" ");
                for (int i = 0; i < originalWords.length; i++) {
                    if (!originalWords[i].equals(correctedWords[i])) {
                        changes++;
                    }
                }
            }

            changesMap.put(sentence, changes);
        }

        List<Map.Entry<String, Integer>> sortedList = new ArrayList<>(changesMap.entrySet());
        sortedList.sort(Map.Entry.comparingByValue());

        for (Map.Entry<String, Integer> entry : sortedList) {
            result.append("    " + entry.getKey() + " | Changes: " + entry.getValue() + "\n");
        }
    }

    private void runCorrector(String input, boolean isFile) {
        try {
            String content;
            if (isFile) 
                content= new String(Files.readAllBytes(Paths.get(input)));
            else
                content = input;
            Corrector corrector = new Corrector();
            String[] sentences = TextProcessor.extractSentences(content).toArray(new String[0]);

            StringBuilder result = new StringBuilder();
            for (String sentence : sentences) {
                sentence = sentence.replaceAll("[^a-zA-Z0-9\\s]", "");
                String[] corrected = corrector.correct(sentence);
                result.append(sentence + " | Corrected Sentence Suggestions:\n");
                printSentencesInOrderOfChanges(corrected, sentence, result);
            }

            resultArea.setText(result.toString());
        } catch (Exception e) {
            resultArea.setText("Error: " + e.getMessage());
        }
    }

    public static void main(String[] args) {
        SwingUtilities.invokeLater(() -> {
            MainApp mainApp = new MainApp();
            mainApp.setVisible(true);
        });
    }
}
